# Importing the JIRA class from the jira module.
import io
import os
from typing import Any, Dict, List, Optional, Tuple, Union

from jira import JIRA  # type: ignore
from jira.client import ResultList  # type: ignore
from jira.resources import Board, Issue  # type: ignore

from pydantic import BaseModel


class Attachment(BaseModel):
    filename: str = ""
    data: Any = io.BytesIO(b"")
    filepath: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True


# This class is a wrapper for the JIRA class in the jira module.
# It adds a few methods to the JIRA class to make it easier
# make it more specialized for the ERP project
class JiraTask(JIRA):
    def __init__(
        self,
        server: str,
        email: str,
        token: str,
        board_name,
        project_key,
        final_stage="Done",
        **kwargs: dict,
    ):
        """The function __init__ is a constructor that takes three parameters.

        Args:
          server (str): The URL of your Jira instance.
          email (str): Email address associated with token.
          token (str): The API token for jira.
          kwargs(dict): Other options defined in JIRA object.
        """
        super(JiraTask, self).__init__(server=server, basic_auth=(email, token))
        self.board_name = board_name
        self.project_key = project_key
        # Used to know whether we should create a new bug report
        self.final_stage = final_stage

    def boards(self, **kwargs) -> ResultList[Board]:
        kwargs["name"] = self.board_name
        """Get a list of board resources.
        Name will be overidden by self.board_name

        Args:
            startAt: The starting index of the returned boards.
                Base index: 0.
            maxResults: The maximum number of boards to return per page.
                Default: 50
            type: Filters results to boards of the specified type.
                Valid values: scrum, kanban.
            name: Filters results to boards that match or partially
                match the specified name.
            projectKeyOrID: Filters results to boards that match
                the specified project key or ID.

        Returns:
            ResultList[Board]
        """
        return super(JiraTask, self).boards(**kwargs)

    @property
    def board_id(self) -> int:
        """
        It returns the id of the first board in the list of boards.

        Returns:
          The id of the first board in the list of boards.
        """
        return self.boards()[0].id

    @property
    def project_id(self) -> int:
        return self.project(self.project_key).id

    def search_issues(
        self, jql_str: str = "", page: int = 1, **kwargs: dict
    ) -> Union[Dict[str, Any], ResultList[Issue]]:
        """
        It searches for issues in Jira.

        Args:
          jql_str: The JQL query string.
          page (int): int = 1. Defaults to 1

        Returns:
          A list of issues
        """
        assert page >= 1
        # convert to zero index
        page = page - 1
        max_results: int = kwargs.get("maxResults", 50)
        kwargs["startAt"] = page * max_results
        return super(JiraTask, self).search_issues(jql_str, **kwargs)

    def does_bug_report_exist(self, title) -> Tuple[bool, Optional[Issue]]:
        search = self.search_issues(
            f'Summary ~ "{title}" and Status != {self.final_stage}'
        )
        if search:
            return True, search.pop()
        return False, None

    def create_bug_report(
        self,
        summary: str,
        description: str,
        attachments: Optional[List[Attachment]] = None,
    ) -> Issue:
        """It creates a bug report in Jira.
        If the bug does not exist it creates a new bug report
        otherwise it makes a comment

        Args:
          summary (str): The title of the bug report.
          description (str): The description of the bug.
          attachments: Optional list of Attachment.

        Returns:
          An issue object
        """
        exist, issue = self.does_bug_report_exist(summary)
        if exist:
            self.add_comment(str(issue), description)
        else:
            assert len(summary) > 0
            issue_dict = {
                "project": {"id": self.project_id},
                "summary": summary,
                "description": description,
                "issuetype": {"name": "Bug"},
            }
            issue = self.create_issue(fields=issue_dict)
        if attachments is None:
            attachments = []
        if attachments:
            existing_attachments = {
                existing_atatchment.filename
                for existing_atatchment in issue.get_field("attachment")
            }
            for attachment in attachments:
                if attachment.filename not in existing_attachments:
                    self.add_attachment(
                        issue=issue,
                        attachment=attachment.data,
                        filename=attachment.filename,
                    )
        return issue


def create_connection():
    email = os.environ.get("jira_email")
    token = os.environ.get("jira_token")
    server = os.environ.get("jira_server")
    board_name = os.environ.get("jira_board_name")
    project_key = os.environ.get("jira_project_key")
    # all environment variables must be set
    assert all((email, token, server, board_name, project_key))
    return JiraTask(server, email, token, board_name, project_key)



"""Exception Handler for DRF sending bug reports to Jira."""
import datetime
import os
import sys
from typing import Iterable, Optional

from jinja2 import (  # type: ignore
    Environment,
    FileSystemLoader,
)

from rest_framework.request import Request  # type: ignore
from rest_framework.views import exception_handler  # type: ignore

from django_jira.tasks.tasks import Attachment, create_connection  # type: ignore


def get_attachments(files: Iterable):
    """
    It takes a list of files and returns a list of attachments

    Args:
      files: A list of files to attach to the email.

    Returns:
      A list of Attachment objects.
    """
    return [Attachment(filename=f.name, data=f) for f in files]


def trace_error(exc: Exception):
    """Get Information on error.

    It takes an exception, and returns a dictionary with the exception's type,
    message, traceback, and any files attached to
    the request

    Args:
      exc (Exception): The exception that was raised.

    Returns:
      A dictionary with the type of error, the message,
        the trace, and the attachments.
    """
    trace = []
    tb = exc.__traceback__
    files: set = set()
    while tb is not None:
        locals = tb.tb_frame.f_locals
        request: Optional[Request] = locals.get("request")
        try:
            username: str = request.user.username  # type: ignore
        except AttributeError:
            username = ""
        try:
            for _, value in request.FILES.items():  # type: ignore
                files.add(value)
        except AttributeError:
            pass

        f_code = tb.tb_frame.f_code
        trace.append(
            {
                "filename": f_code.co_filename,
                "name": f_code.co_name,
                "lineno": tb.tb_lineno,
                "locals": locals,
                "request": request,
                "username": username,
            }
        )
        tb = tb.tb_next

    return {
        "type": type(exc).__name__,
        "message": str(exc),
        "trace": trace,
        "attachments": get_attachments(files),
    }


def error_title(trace: dict):
    """Title for Issue in Jira.
    It takes a traceback dictionary and returns a string
        that is the title of the error.

    Args:
      trace (dict): The traceback object.

    Returns:
      A string with the title of the project, the type of error,
      the error message, the filename, and the line number.
    """
    return (
        f'{os.environ.get("PROJECT_NAME", "Unknown Project").title()} '
        f'{trace["type"]}: '
        f'{trace["message"].title()} in '
        f'{trace["trace"][-1]["filename"]} at line '
        f'{trace["trace"][-1]["lineno"]}'.replace("\\", "/")
    )


def drf_jira_exception_handler(exc: Exception, context: dict):
    """Custom Exception Handler creates Jira Bug Reports.

    Returns the same as
    creates a JIRA bug report, and returns a response to the user

    Args:
      exc (Exception): The exception instance raised.
      context (dict): The context parameter is a
        dictionary that contains the following keys:


    Returns:
      The response is being returned.
    """
    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = exception_handler(exc, context)
    if response is None or response.status_code >= 500:
        template_path = (
            os.path.dirname(
                str(sys.modules[drf_jira_exception_handler.__module__].__file__)
            )
            + "/templates"
        )
        environment = Environment(
            loader=FileSystemLoader(searchpath=template_path), autoescape=True
        )
        template = environment.get_template("drf_error_template.md")
        trace = trace_error(exc)

        content = template.render(
            exc=exc,
            context=context,
            exc_name=error_title(trace),
            request=context["request"],
            parameters=zip(
                context["request"].query_params.keys(),
                context["request"].query_params.values(),
            ),
            time=datetime.datetime.now(),
            stacktrace=trace["trace"],
        )
        # content = str(trace_error(exc))
        jira = create_connection()
        jira.create_bug_report(error_title(trace), content, trace["attachments"])

    return response



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['django_jira',
 'django_jira.error_handler',
 'django_jira.error_handler.templates',
 'django_jira.tasks']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2.0,<4.0.0',
 'Jinja2>=3.1.2,<4.0.0',
 'djangorestframework>=3.13.1,<4.0.0',
 'jira>=3.4.1,<4.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'python-dotenv>=0.21.0,<0.22.0']

setup_kwargs = {
    'name': 'django-jira',
    'version': '0.2.0',
    'description': 'Intergrates Jira library into django microservices',
    'long_description': '# Django-Jira\n\nAllows for creating issues in Jira with Django RestFramework\n',
    'author': 'Antoine Wood',
    'author_email': 'antoinewood@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
